"""Image crawler for ShopStyle on aws dynamodb.
"""

import datetime
import json
import os
import requests
import time
import urllib

import multiprocessing
from tqdm import tqdm

import objectpath

import owl
from owl.data import img_tools
from owl.third.aws import dynamodb, s3

# create logger.
logger_name = "crawler_logger"
if not os.path.exists("extra"):
  os.mkdir("extra")
logger = owl.data.common.get_logger(logger_name,
                                    os.path.join("extra",
                                                 logger_name + ".log"))


class ShopStyleParams(object):
  """Parameters for shopstyle api.
  """
  api_name = "ShopStyle"
  base_url = "http://api.shopstyle.com/"
  product_url = "api/v2/products"
  pid = "uid5625-32673475-18"
  format = "json"
  # keywords
  fts = ""
  cat = "category"
  offset = "0"
  page_limit = "50"
  results_num = 100000
  retailer_id = ""
  img_save_dir = ""
  meta_save_dir = ""
  source_name = ""
  # db or file.
  save_type = "db"

  def get_request_url(self):
    """Form url for request.
    """
    return self.base_url + self.product_url + '?pid=' + self.pid + \
        '&cat=' + self.cat + '&format=' + self.format + \
        '&offset=' + self.offset + '&limit=' + self.page_limit + "&fl=" + self.retailer_id


class ShopStyleCrawlerAWS(object):
  """Crawler for shopstyle.

  Metadata is saved to dynamodb.
  Images are saved to s3.
  """
  db_config = None
  dynamodb_man = None
  s3_man = None
  table_name = "eyestyle.eloquii.shopstyle"

  def init(self):
    """Initialize database.

    Load configuration and create/get table.
    """
    try:
      self.db_config = json.load(open("./configs/aws_db_config.json", "r"))
      self.dynamodb_man = dynamodb.DynamoDBIO(
          access_key=self.db_config["access_key"],
          secret_key=self.db_config["secret_access_key"],
          region=self.db_config["region"])
      self.s3_man = s3.S3IO(
          access_key=self.db_config["access_key"],
          secret_key=self.db_config["secret_access_key"],
          region=self.db_config["region"])
      self.s3_man.use_bucket(bucket_name=self.db_config["s3_bucket"])
      logger.info("using s3 bucket: {}".format(self.db_config["s3_bucket"]))
      self.dynamodb_man.create_table(self.table_name, "item_id", None, 10, 10)
    except Exception as ex:
      logger.error("init db error: {}".format(ex.message))
      raise

  def save_item_to_db(self, original_id, img_url, item_metadata, sparams):
    """Save item to dynamodb.

    Args:
      original_id: id from the data source.
      img_url: product img url.
      sparams: parameters.
      item_metadata: metadata for the item.
    """
    try:
      item_id = "{}___shopstyle".format(original_id)

      # check if already exists.
      res = self.dynamodb_man.get_item({"item_id": item_id})
      if res is not None:
        return

      # req_data = requests.get(product_img_url, headers=header)
      startt = time.time()
      img_bin, img_ext = img_tools.download_img_from_url(img_url)
      saved_img_url = self.s3_man.upload_img_bin_as_file(
          self.db_config["s3_bucket"], img_bin, img_ext)
      print("save image time: {}".format(time.time() - startt))

      # add to db.
      save_date = datetime.datetime.now().isoformat()
      target_item = {
          "item_id": item_id,
          "original_id": original_id,
          "data_source": "shopstyle",
          "date": save_date,
          "category": item_metadata["categories"][0]["id"],
          "img_url": saved_img_url,
          "img_width": -1,
          "img_height": -1,
          "img_format": img_ext,
          "in_stock": item_metadata["inStock"],
          "raw_metadata": json.dumps(item_metadata)
      }
      target_item["brand"] = objectpath.Tree(item_metadata).execute(
          "$.brand.name")

      startt = time.time()
      self.dynamodb_man.add_item(target_item)
      print("save dynamodb time: {}s".format(time.time() - startt))
    except Exception as ex:
      logger.error("error in saving to db: {}".format(ex.message))

  def retrieve_items(self, items, sparams):
    """Retrieve items from api response.

    Args:
      items: product items to save.
      sparams: request parameters.
    """
    # create db item.
    for item in tqdm(items):
      try:
        #print('processing {}th product'.format(cnt))
        product_id = str(item["id"])
        use_color = False
        if "color" in item:
          # save all color versions.
          for color_item in item["colors"]:
            if "image" not in color_item:
              continue
            product_img_id = color_item["image"]["id"]
            original_id = "{}__{}".format(product_id, product_img_id)
            product_img_url = color_item["image"]["sizes"]["Original"]["url"]
            self.save_item_to_db(original_id, product_img_url, item, sparams)
            use_color = True

        # just use one image.
        if not use_color:
          if "image" not in item:
            return
          product_img_id = item["image"]["id"]
          original_id = "{}__{}".format(product_id, product_img_id)
          product_img_url = item["image"]["sizes"]["Original"]["url"]
          self.save_item_to_db(original_id, product_img_url, item, sparams)

      except Exception as e:
        logger.error("error in shopstyle crawler: {}".format(e.message))

  def crawl(self, params):
    """Crawl products from api.

    Args:
      params: params for crawling.
    """
    try:
      # get total number.
      params.offset = "0"
      url = params.get_request_url()
      res_data = requests.get(url)
      res_json = json.loads(res_data.content)
      total_num = int(res_json["metadata"]["total"])
      total_num = min(params.results_num, total_num)
      logger.info("category {} has total {} products.".format(
          params.cat, total_num))
      page_num = max(1, total_num / int(params.page_limit))
      logger.info("total page number to fetch: {}".format(page_num))

      # make request for each page.
      for pid in range(page_num):
        try:
          logger.info("making api call for {} at page {}:{}".format(
              params.cat, pid + 1, page_num))
          cur_offset = int(params.page_limit) * pid
          if cur_offset > total_num:
            break
          params.offset = str(cur_offset)
          url = params.get_request_url()
          # make request.
          res_data = requests.get(url)
          res_json = json.loads(res_data.content)
        except Exception as ex:
          logger.error("error in response: {}".format(ex.message))
          continue

        # parse results.
        if len(res_json["products"]) == 0:
          logger.warning("empty response")
        else:
          self.retrieve_items(res_json["products"], params)

        # parse for api call
        time.sleep(3)
    except Exception as ex:
      logger.error("error in crawl: {}".format(ex.message))


def crawl_one_cate_per_process(cate):
  """Crawl one category in a standalone process.

    Args:
      cate: category name.
    """
  try:
    crawler = ShopStyleCrawlerAWS()
    crawler.init()
    sparams = ShopStyleParams()
    sparams.cat = cate
    crawler.crawl(sparams)
  except Exception as ex:
    logger.error("error crawl {}: ".format(cate, ex.messeage))


def run_crawler_parallel():
  """Run crawler in multiple processes.
  """
  try:
    logger.info("parallel crawler started...")
    # get category list.
    cate_fn = "./shopstyle_categories.json"
    cate_data = json.load(open(cate_fn, 'r'))
    cates = []
    for gender in cate_data:
      for cate in cate_data[gender]:
        for subcate in cate["sub"]:
          cates.append(subcate["id"])
    # start processing.
    num_cores = multiprocessing.cpu_count()
    pool = multiprocessing.Pool(processes=num_cores)
    pool.map(crawl_one_cate_per_process, cates)
    logger.info("completed crawl.")
  except Exception as ex:
    logger.error("error in main parallel crawler: {}".format(ex.message))


def run_crawler_single():
  """Run crawler in single process.
  """
  try:
    logger.info("crawler started...")
    crawler = ShopStyleCrawlerAWS()
    crawler.init()
    sparams = ShopStyleParams()
    cate_fn = "shopstyle_categories.json"
    cate_data = json.load(open(cate_fn, 'r'))
    cates = []
    for gender in cate_data:
      for cate in cate_data[gender]:
        for subcate in cate["sub"]:
          cates.append(subcate["id"])
    for cate in cates:
      sparams.cat = cate
      crawler.crawl(sparams)
    logger.info("completed crawl.")
  except Exception as ex:
    logger.error("main exception: {}".format(ex.message))


def run_crawler_retailer(retailer):
  """Run crawler in single process.
  """
  try:
    logger.info("crawler started...")
    crawler = ShopStyleCrawlerAWS()
    crawler.init()
    sparams = ShopStyleParams()
    sparams.retailer_id = retailer
    crawler.crawl(sparams)
    logger.info("completed crawl.")
  except Exception as ex:
    logger.error("main exception: {}".format(ex.message))


if __name__ == "__main__":
  # while True:
  #   run_crawler_parallel()
  #   time.sleep(10)
  run_crawler_retailer("r2001")


"""Linkedin Data API.

https://developer.linkedin.com/docs/guide/v2
"""


class LinkedInUserProfile(object):
  """Class for member profile.
  """
  pass


class LinkedInAPI(object):
  """Class for accessing Linkedin data.
  """
  _client_id = None
  _secret_id = None
  base_url = "https://api.linkedin.com/v1/people"

  def __init__(self, client_id, secret_id):
    self._client_id = client_id
    self._secret_id = secret_id


if __name__ == "__main__":
  pass


"""Tests for unsplash api.
"""

import json

from datahunters.stock.unsplash.api import UnsplashAPI, UnsplashImage


class TestUnsplashAPI(object):
  """Class for testing api.
  """
  engine = None

  def init(self):
    with open("configs.json", "r") as f:
      configs = json.load(f)
    self.engine = UnsplashAPI(configs["api_key"])

  def test_get_imgs(self):
    self.init()
    imgs = self.engine.get_imgs(0, 30)
    assert len(imgs) == 30

  def test_search_imgs(self):
    self.init()
    imgs = self.engine.search_imgs("people")
    assert len(imgs) > 0

  def test_random_imgs(self):
    self.init()
    imgs = self.engine.get_random_imgs()
    assert len(imgs) > 0



"""TMDB api wrapper.
"""

import json
import requests

from datahunters.movies import common


class TMDBAPI(common.MovieAPIBase):
  """Class for TMDB.
  """
  engine = None
  api_key = ""
  root_url = "https://api.themoviedb.org/3/"

  def __init__(self, api_key):
    super(TMDBAPI, self).__init__(api_key)
    with open("configs.json", "r") as f:
      configs = json.load(f)
      self.api_key = configs["api_key"]
    print("TMDB api init")

  def get_movie_info(self, id):
    movie_url = "{}movie/{}?api_key={}&language=en-US".format(
        self.root_url, id, self.api_key)
    res = requests.request("GET", movie_url, data={})
    movie_data = self.convert_to_movieobj(res.json())
    return movie_data

  def map_to_internal_genre(self, genre):
    if genre == "Science Fiction":
      return "Sci-Fi"
    return genre

  def convert_to_movieobj(self, tmdb_movie):
    """Convert TMDB movie object to internal movie object.
    """
    inter_movie = common.MovieObject()
    inter_movie.title = tmdb_movie["title"]
    inter_movie.release_date = tmdb_movie["release_date"]
    inter_movie.imdb_id = tmdb_movie["imdb_id"]
    inter_movie.desc = tmdb_movie["overview"]
    return inter_movie

  def get_person_info(self, id):
    pass


if __name__ == "__main__":
  api = TMDBAPI(None)
  data = api.get_movie_info("10195")
  print(data.title)



"""Tools for social data hunters.
"""

import re

import robobrowser


def gen_fb_token(fb_email, fb_pw):
  """Retrieve facebook token using standard credentials.
  """
  MOBILE_USER_AGENT = "Mozilla/5.0 (Linux; U; en-gb; KFTHWI Build/JDQ39) AppleWebKit/535.19 (KHTML, like Gecko) Silk/3.16 Safari/535.19"
  FB_AUTH = "https://www.facebook.com/v2.6/dialog/oauth?redirect_uri=fb464891386855067%3A%2F%2Fauthorize%2F&display=touch&state=%7B%22challenge%22%3A%22IUUkEUqIGud332lfu%252BMJhxL4Wlc%253D%22%2C%220_auth_logger_id%22%3A%2230F06532-A1B9-4B10-BB28-B29956C71AB1%22%2C%22com.facebook.sdk_client_state%22%3Atrue%2C%223_method%22%3A%22sfvc_auth%22%7D&scope=user_birthday%2Cuser_photos%2Cuser_education_history%2Cemail%2Cuser_relationship_details%2Cuser_friends%2Cuser_work_history%2Cuser_likes&response_type=token%2Csigned_request&default_audience=friends&return_scopes=true&auth_type=rerequest&client_id=464891386855067&ret=login&sdk=ios&logger_id=30F06532-A1B9-4B10-BB28-B29956C71AB1&ext=1470840777&hash=AeZqkIcf-NEW6vBd"
  s = robobrowser.RoboBrowser(user_agent=MOBILE_USER_AGENT, parser="lxml")
  s.open(FB_AUTH)
  f = s.get_form()
  f["pass"] = fb_pw
  f["email"] = fb_email
  s.submit_form(f)
  f = s.get_form()
  s.submit_form(f, submit=f.submit_fields['__CONFIRM__'])
  access_token = re.search(r"access_token=([\w\d]+)",
                           s.response.content.decode()).groups()[0]
  return access_token


def gen_fb_token_nonexpired(fb_app_id, fb_app_secret):
  """Use pipe to generate never expire token.
  """
  return "{}|{}".format(fb_app_id, fb_app_secret)


def unicode_normalize(text):
  """Transform unicode.
  """
  return text.translate({
      0x2018: 0x27,
      0x2019: 0x27,
      0x201C: 0x22,
      0x201D: 0x22,
      0xa0: 0x20
  }).encode('utf-8')


if __name__ == "__main__":
  pass



"""Tests for Adobe Stock API.
"""

import json

from datahunters.stock.adobe.api import AdobeStockAPI


class TestAdobeStockAPI(object):
  """Class for tests.
  """

  def init(self):
    with open("configs.json", "r") as f:
      configs = json.load(f)
    self.api = AdobeStockAPI(configs["api_key"])

  def test_get_imgs(self):
    self.init()
    img_gen2 = self.api.get_imgs()
    assert next(img_gen2).img_id != ""

  def test_search_imgs(self):
    self.init()
    img_gen = self.api.search_imgs("cats in hat")
    assert next(img_gen).img_id != ""



"""Google api class.
"""

import urllib

from datahunters.shared.selenium_scraper import SeleniumScraper
from datahunters.search.common import ImgSearchObject


class GoogleImageAPI(SeleniumScraper):
  """Class for google image search api.
  """
  base_url = "https://www.google.com/search"

  def __init__(self, use_headless=True):
    super().__init__(use_headless)
    print("Google image api initialized")

  def convert_elem_to_obj(self, elem):
    """Convert page element to image object.
    """
    cur_res = ImgSearchObject()
    link = self.get_attribute_values([elem], "href")[0]
    link = urllib.parse.unquote(link)
    cur_res.img_url = link[link.find("imgurl=") + 7:link.find(
        "&imgrefurl")].rstrip(".")
    return cur_res

  def scrape(self, keywords, get_all=False):
    """Collect images from google search results.

    Args:
      keywords: search input.
      item_num: number of items to retrieve.
      offset: offset for results to start.

    Returns:
      list of image items.
    """
    all_res = []
    try:
      print("start scraping '{}' using Google".format(keywords))
      formatted_keywords = keywords.strip().replace(" ", "+")
      req_url = "{}?q={}&source=lnms&tbm=isch&sa=X&ei=0eZEVbj3IJG5uATalICQAQ&ved=0CAcQ_AUoAQ&biw=939&bih=591".format(
          self.base_url, formatted_keywords)
      # check default page.
      print("checking default data")
      if not get_all:
        self.load_content(req_url)
        elems = self.find_elements("a.rg_l")
      else:
        elems = self.scrape_inf_scroll(
            req_url, "a.rg_l", None, load_btn_selector="input#smb")
      print("total fetched items: {}".format(len(elems)))
      for elem in elems:
        try:
          cur_res = self.convert_elem_to_obj(elem)
          all_res.append(cur_res)
        except Exception as ex:
          print("error in processing item: {}".format(ex))
          continue
      print("Google image scraping finished.")
      return all_res
    except Exception as ex:
      print("error in Google image scraper: {}".format(ex))
      return all_res



#!/usr/bin/python
# -*- coding: utf-8 -*-

import argparse
import codecs
import errno
import glob
from operator import itemgetter
import json
import logging.config
import os
import re
import sys
import textwrap
import time
import warnings

import concurrent.futures
import requests
import tqdm

from instagram_scraper.constants import *

try:
  reload(sys)  # Python 2.7
  sys.setdefaultencoding("UTF8")
except NameError:
  pass

warnings.filterwarnings('ignore')

BASE_URL = 'https://www.instagram.com/'
LOGIN_URL = BASE_URL + 'accounts/login/ajax/'
LOGOUT_URL = BASE_URL + 'accounts/logout/'
CHROME_WIN_UA = 'Mozilla/5.0 (Windows NT 10.0; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/59.0.3071.115 Safari/537.36'
USER_URL = BASE_URL + '{0}/?__a=1'

STORIES_URL = 'https://i.instagram.com/api/v1/feed/user/{0}/reel_media/'
STORIES_UA = 'Instagram 9.5.2 (iPhone7,2; iPhone OS 9_3_3; en_US; en-US; scale=2.00; 750x1334) AppleWebKit/420+'
STORIES_COOKIE = 'ds_user_id={0}; sessionid={1};'

TAGS_URL = BASE_URL + 'explore/tags/{0}/?__a=1'
LOCATIONS_URL = BASE_URL + 'explore/locations/{0}/?__a=1'
VIEW_MEDIA_URL = BASE_URL + 'p/{0}/?__a=1'
SEARCH_URL = BASE_URL + 'web/search/topsearch/?context=blended&query={0}'

QUERY_COMMENTS = BASE_URL + 'graphql/query/?query_id=17852405266163336&shortcode={0}&first=100&after={1}'
QUERY_HASHTAG = BASE_URL + 'graphql/query/?query_id=17882293912014529&tag_name={0}&first=100&after={1}'
QUERY_LOCATION = BASE_URL + 'graphql/query/?query_id=17881432870018455&id={0}&first=100&after={1}'
QUERY_MEDIA = BASE_URL + 'graphql/query/?query_id=17888483320059182&id={0}&first=100&after={1}'


class InstagramScraper(object):
  """InstagramScraper scrapes and downloads an instagram user's photos and videos"""

  def __init__(self, **kwargs):
    default_attr = dict(
        username='',
        usernames=[],
        filename=None,
        login_user=None,
        login_pass=None,
        login_only=False,
        destination='./',
        retain_username=False,
        quiet=False,
        maximum=0,
        media_metadata=False,
        latest=False,
        media_types=['image', 'video', 'story'],
        tag=False,
        location=False,
        search_location=False,
        comments=False,
        verbose=0,
        include_location=False,
        filter=None)

    allowed_attr = list(default_attr.keys())
    default_attr.update(kwargs)

    for key in default_attr:
      if key in allowed_attr:
        self.__dict__[key] = kwargs.get(key)

    # Set up a logger
    self.logger = InstagramScraper.get_logger(
        level=logging.DEBUG, verbose=default_attr.get('verbose'))

    self.posts = []
    self.session = requests.Session()
    self.session.headers = {'user-agent': CHROME_WIN_UA}

    self.cookies = None
    self.logged_in = False
    self.last_scraped_filemtime = 0
    if default_attr['filter']:
      self.filter = list(self.filter)

  def login(self):
    """Logs in to instagram."""
    self.session.headers.update({'Referer': BASE_URL})
    req = self.session.get(BASE_URL)

    self.session.headers.update({'X-CSRFToken': req.cookies['csrftoken']})

    login_data = {'username': self.login_user, 'password': self.login_pass}
    login = self.session.post(LOGIN_URL, data=login_data, allow_redirects=True)
    self.session.headers.update({'X-CSRFToken': login.cookies['csrftoken']})
    self.cookies = login.cookies
    login_text = json.loads(login.text)

    if login_text.get('authenticated') and login.status_code == 200:
      self.logged_in = True
    else:
      self.logger.error('Login failed for ' + self.login_user)

      if 'checkpoint_url' in login_text:
        self.logger.error('Please verify your account at ' + login_text.get(
            'checkpoint_url'))
      elif 'errors' in login_text:
        for count, error in enumerate(login_text['errors'].get('error')):
          count += 1
          self.logger.debug('Session error %(count)s: "%(error)s"' % locals())
      else:
        self.logger.error(json.dumps(login_text))

  def logout(self):
    """Logs out of instagram."""
    if self.logged_in:
      try:
        logout_data = {'csrfmiddlewaretoken': self.cookies['csrftoken']}
        self.session.post(LOGOUT_URL, data=logout_data)
        self.logged_in = False
      except requests.exceptions.RequestException:
        self.logger.warning('Failed to log out ' + self.login_user)

  def make_dst_dir(self, username):
    """Creates the destination directory."""
    if self.destination == './':
      dst = './' + username
    else:
      if self.retain_username:
        dst = self.destination + '/' + username
      else:
        dst = self.destination

    try:
      os.makedirs(dst)
    except OSError as err:
      if err.errno == errno.EEXIST and os.path.isdir(dst):
        # Directory already exists
        self.get_last_scraped_filemtime(dst)
        pass
      else:
        # Target dir exists as a file, or a different error
        raise

    return dst

  def get_last_scraped_filemtime(self, dst):
    """Stores the last modified time of newest file in a directory."""
    list_of_files = []
    file_types = ('*.jpg', '*.mp4')

    for type in file_types:
      list_of_files.extend(glob.glob(dst + '/' + type))

    if list_of_files:
      latest_file = max(list_of_files, key=os.path.getmtime)
      self.last_scraped_filemtime = int(os.path.getmtime(latest_file))

  def query_comments_gen(self, shortcode, end_cursor=''):
    """Generator for comments."""
    comments, end_cursor = self.__query_comments(shortcode, end_cursor)

    if comments:
      try:
        while True:
          for item in comments:
            yield item

          if end_cursor:
            comments, end_cursor = self.__query_comments(shortcode, end_cursor)
          else:
            return
      except ValueError:
        self.logger.exception('Failed to query comments for shortcode ' +
                              shortcode)

  def __query_comments(self, shortcode, end_cursor=''):
    resp = self.session.get(QUERY_COMMENTS.format(shortcode, end_cursor))

    if resp.status_code == 200:
      payload = json.loads(resp.text)['data']['shortcode_media']

      if payload:
        container = payload['edge_media_to_comment']
        comments = [node['node'] for node in container['edges']]
        end_cursor = container['page_info']['end_cursor']
        return comments, end_cursor
      else:
        return iter([])
    else:
      time.sleep(6)
      return self.__query_comments(shortcode, end_cursor)

  def scrape_hashtag(self):
    self.__scrape_query(self.query_hashtag_gen)

  def scrape_location(self):
    self.__scrape_query(self.query_location_gen)

  def __scrape_query(
      self,
      media_generator,
      executor=concurrent.futures.ThreadPoolExecutor(max_workers=10)):
    """Scrapes the specified value for posted media."""
    for value in self.usernames:
      self.posts = []
      self.last_scraped_filemtime = 0
      future_to_item = {}

      dst = self.make_dst_dir(value)

      if self.include_location:
        media_exec = concurrent.futures.ThreadPoolExecutor(max_workers=5)

      iter = 0
      for item in tqdm.tqdm(
          media_generator(value),
          desc='Searching {0} for posts'.format(value),
          unit=" media",
          disable=self.quiet):
        if ((item['is_video'] is False and 'image' in self.media_types) or \
                    (item['is_video'] is True and 'video' in self.media_types)
            ) and self.is_new_media(item):
          future = executor.submit(self.download, item, dst)
          future_to_item[future] = item

        if self.include_location and 'location' not in item:
          media_exec.submit(self.__get_location, item)

        if self.comments:
          item['edge_media_to_comment']['data'] = list(
              self.query_comments_gen(item['shortcode']))

        if self.media_metadata or self.comments or self.include_location:
          self.posts.append(item)

        iter = iter + 1
        if self.maximum != 0 and iter >= self.maximum:
          break

      if future_to_item:
        for future in tqdm.tqdm(
            concurrent.futures.as_completed(future_to_item),
            total=len(future_to_item),
            desc='Downloading',
            disable=self.quiet):
          item = future_to_item[future]

          if future.exception() is not None:
            self.logger.warning(
                'Media for {0} at {1} generated an exception: {2}'.format(
                    value, item['urls'], future.exception()))

      if (self.media_metadata or self.comments or
          self.include_location) and self.posts:
        self.save_json(self.posts, '{0}/{1}.json'.format(dst, value))

  def query_hashtag_gen(self, hashtag):
    return self.__query_gen(QUERY_HASHTAG, 'hashtag', hashtag)

  def query_location_gen(self, location):
    return self.__query_gen(QUERY_LOCATION, 'location', location)

  def __query_gen(self, url, entity_name, query, end_cursor=''):
    """Generator for hashtag and location."""
    nodes, end_cursor = self.__query(url, entity_name, query, end_cursor)

    if nodes:
      try:
        while True:
          for node in nodes:
            yield node

          if end_cursor:
            nodes, end_cursor = self.__query(url, entity_name, query,
                                             end_cursor)
          else:
            return
      except ValueError:
        self.logger.exception('Failed to query ' + query)

  def __query(self, url, entity_name, query, end_cursor):
    resp = self.session.get(url.format(query, end_cursor))

    if resp.status_code == 200:
      payload = json.loads(resp.text)['data'][entity_name]

      if payload:
        nodes = []

        if end_cursor == '':
          top_posts = payload['edge_' + entity_name + '_to_top_posts']
          nodes.extend(self._get_nodes(top_posts))

        posts = payload['edge_' + entity_name + '_to_media']

        nodes.extend(self._get_nodes(posts))
        end_cursor = posts['page_info']['end_cursor']
        return nodes, end_cursor
      else:
        return iter([])
    else:
      time.sleep(6)
      return self.__query(url, entity_name, query, end_cursor)

  def _get_nodes(self, container):
    return [self.augment_node(node['node']) for node in container['edges']]

  def augment_node(self, node):
    self.extract_tags(node)

    details = None
    if self.include_location and 'location' not in node:
      details = self.__get_media_details(node['shortcode'])
      node['location'] = details.get('location') if details else None

    if 'urls' not in node:
      node['urls'] = []

    if node['is_video'] and 'video_url' in node:
      node['urls'] = [node['video_url']]
    elif '__typename' in node and node['__typename'] == 'GraphImage':
      node['urls'] = [self.get_original_image(node['display_url'])]
    else:
      if details is None:
        details = self.__get_media_details(node['shortcode'])

      if details:
        if '__typename' in details and details['__typename'] == 'GraphVideo':
          node['urls'] = [details['video_url']]
        elif '__typename' in details and details[
            '__typename'] == 'GraphSidecar':
          urls = []
          for carousel_item in details['edge_sidecar_to_children']['edges']:
            urls += self.augment_node(carousel_item['node'])['urls']
          node['urls'] = urls
        else:
          node['urls'] = [self.get_original_image(details['display_url'])]

    return node

  def __get_media_details(self, shortcode):
    resp = self.session.get(VIEW_MEDIA_URL.format(shortcode))

    if resp.status_code == 200:
      try:
        return json.loads(resp.text)['graphql']['shortcode_media']
      except ValueError:
        self.logger.warning('Failed to get media details for ' + shortcode)

    else:
      self.logger.warning('Failed to get media details for ' + shortcode)

  def __get_location(self, item):
    code = item.get('shortcode', item.get('code'))

    if code:
      details = self.__get_media_details(code)
      item['location'] = details.get('location')

  def scrape(self,
             executor=concurrent.futures.ThreadPoolExecutor(max_workers=10)):
    """Crawls through and downloads user's media"""
    if self.login_user and self.login_pass:
      self.login()
      if not self.logged_in and self.login_only:
        self.logger.warning('Fallback anonymous scraping disabled')
        return

    for username in self.usernames:
      self.posts = []
      self.last_scraped_filemtime = 0
      future_to_item = {}

      dst = self.make_dst_dir(username)

      # Get the user metadata.
      user = self.fetch_user(username)

      if user:
        self.get_profile_pic(dst, executor, future_to_item, user, username)
        self.get_stories(dst, executor, future_to_item, user, username)

      # Crawls the media and sends it to the executor.
      try:
        user = self.get_user(username)

        if not user:
          continue

        self.get_media(dst, executor, future_to_item, user)

        # Displays the progress bar of completed downloads. Might not even pop up if all media is downloaded while
        # the above loop finishes.
        if future_to_item:
          for future in tqdm.tqdm(
              concurrent.futures.as_completed(future_to_item),
              total=len(future_to_item),
              desc='Downloading',
              disable=self.quiet):
            item = future_to_item[future]

            if future.exception() is not None:
              self.logger.warning('Media at {0} generated an exception: {1}'.
                                  format(item['urls'], future.exception()))

        if (self.media_metadata or self.comments or
            self.include_location) and self.posts:
          self.save_json(self.posts, '{0}/{1}.json'.format(dst, username))
      except ValueError:
        self.logger.error("Unable to scrape user - %s" % username)
    self.logout()

  def get_profile_pic(self, dst, executor, future_to_item, user, username):
    # Download the profile pic if not the default.
    if 'image' in self.media_types and 'profile_pic_url_hd' in user \
            and '11906329_960233084022564_1448528159' not in user['profile_pic_url_hd']:
      item = {
          'urls':
          [re.sub(r'/[sp]\d{3,}x\d{3,}/', '/', user['profile_pic_url_hd'])],
          'created_time': 1286323200
      }

      if self.latest is False or os.path.isfile(dst + '/' + item['urls'][0]
                                                .split('/')[-1]) is False:
        for item in tqdm.tqdm(
            [item],
            desc='Searching {0} for profile pic'.format(username),
            unit=" images",
            ncols=0,
            disable=self.quiet):
          future = executor.submit(self.download, item, dst)
          future_to_item[future] = item

  def get_stories(self, dst, executor, future_to_item, user, username):
    """Scrapes the user's stories."""
    if self.logged_in and 'story' in self.media_types:
      # Get the user's stories.
      stories = self.fetch_stories(user['id'])

      # Downloads the user's stories and sends it to the executor.
      iter = 0
      for item in tqdm.tqdm(
          stories,
          desc='Searching {0} for stories'.format(username),
          unit=" media",
          disable=self.quiet):
        future = executor.submit(self.download, item, dst)
        future_to_item[future] = item

        iter = iter + 1
        if self.maximum != 0 and iter >= self.maximum:
          break

  def get_user(self, username):
    """Fetches the user's metadata."""
    url = USER_URL.format(username)

    resp = self.session.get(url)

    if resp.status_code == 200:
      user = json.loads(resp.text)['user']
      if user and user['is_private'] and user['media'][
          'count'] > 0 and not user['media']['nodes']:
        self.logger.error('User {0} is private'.format(username))
      return user
    else:
      self.logger.error(
          'Error getting user details for {0}. Please verify that the user exists.'.
          format(username))

  def get_media(self, dst, executor, future_to_item, user):
    """Scrapes the user's posts for media."""
    if self.media_types == ['story']:
      return

    username = user['username']

    if self.include_location:
      media_exec = concurrent.futures.ThreadPoolExecutor(max_workers=5)

    iter = 0
    for item in tqdm.tqdm(
        self.query_media_gen(user),
        desc='Searching {0} for posts'.format(username),
        unit=' media',
        disable=self.quiet):
      # -Filter command line
      if self.filter:
        if 'tags' in item:
          filtered = any(x in item['tags'] for x in self.filter)
          if self.has_selected_media_types(item) and self.is_new_media(
              item) and filtered:
            future = executor.submit(self.download, item, dst)
            future_to_item[future] = item
        else:
          # For when filter is on but media doesnt contain tags
          pass
      # --------------#
      else:
        if self.has_selected_media_types(item) and self.is_new_media(item):
          future = executor.submit(self.download, item, dst)
          future_to_item[future] = item

      if self.include_location:
        media_exec.submit(self.__get_location, item)

      if self.comments:
        item['comments'] = {
            'data': list(self.query_comments_gen(item['shortcode']))
        }

      if self.media_metadata or self.comments or self.include_location:
        self.posts.append(item)

      iter = iter + 1
      if self.maximum != 0 and iter >= self.maximum:
        break

  def fetch_user(self, username):
    """Fetches the user's metadata."""
    resp = self.session.get(BASE_URL + username)

    if resp.status_code == 200 and '_sharedData' in resp.text:
      try:
        shared_data = resp.text.split("window._sharedData = ")[1].split(
            ";</script>")[0]
        return json.loads(shared_data)['entry_data']['ProfilePage'][0]['user']
      except (TypeError, KeyError, IndexError):
        pass

  def fetch_stories(self, user_id):
    """Fetches the user's stories."""
    resp = self.session.get(
        STORIES_URL.format(user_id),
        headers={
            'user-agent': STORIES_UA,
            'cookie': STORIES_COOKIE.format(self.cookies['ds_user_id'],
                                            self.cookies['sessionid'])
        })

    retval = json.loads(resp.text)

    if resp.status_code == 200 and 'items' in retval and len(retval[
        'items']) > 0:
      return [self.set_story_url(item) for item in retval['items']]
    return []

  def query_media_gen(self, user, end_cursor=''):
    """Generator for media."""
    media, end_cursor = self.__query_media(user['id'], end_cursor)

    if media:
      try:
        while True:
          for item in media:
            if self.latest and self.last_scraped_filemtime >= self.__get_timestamp(
                item):
              return
            yield item

          if end_cursor:
            media, end_cursor = self.__query_media(user['id'], end_cursor)
          else:
            return
      except ValueError:
        self.logger.exception('Failed to query media for user ' + user[
            'username'])

  def __query_media(self, id, end_cursor=''):
    resp = self.session.get(QUERY_MEDIA.format(id, end_cursor))

    if resp.status_code == 200:
      payload = json.loads(resp.text)['data']['user']

      if payload:
        container = payload['edge_owner_to_timeline_media']
        nodes = self._get_nodes(container)
        end_cursor = container['page_info']['end_cursor']
        return nodes, end_cursor
      else:
        return iter([])
    else:
      if resp and resp.text:
        self.logger.warning(resp.text)
      time.sleep(6)
      return self.__query_media(id, end_cursor)

  def has_selected_media_types(self, item):
    filetypes = {'jpg': 0, 'mp4': 0}

    for url in item['urls']:
      ext = self.__get_file_ext(url)
      if ext not in filetypes:
        filetypes[ext] = 0
      filetypes[ext] += 1

    if ('image' in self.media_types and filetypes['jpg'] > 0) or \
            ('video' in self.media_types and filetypes['mp4'] > 0):
      return True

    return False

  def extract_tags(self, item):
    """Extracts the hashtags from the caption text."""
    caption_text = ''
    if 'caption' in item and item['caption']:
      if isinstance(item['caption'], dict):
        caption_text = item['caption']['text']
      else:
        caption_text = item['caption']

    elif 'edge_media_to_caption' in item and item[
        'edge_media_to_caption'] and item['edge_media_to_caption']['edges']:
      caption_text = item['edge_media_to_caption']['edges'][0]['node']['text']

    if caption_text:
      # include words and emojis
      item['tags'] = re.findall(
          r"(?<!&)#(\w+|(?:[\xA9\xAE\u203C\u2049\u2122\u2139\u2194-\u2199\u21A9\u21AA\u231A\u231B\u2328\u2388\u23CF\u23E9-\u23F3\u23F8-\u23FA\u24C2\u25AA\u25AB\u25B6\u25C0\u25FB-\u25FE\u2600-\u2604\u260E\u2611\u2614\u2615\u2618\u261D\u2620\u2622\u2623\u2626\u262A\u262E\u262F\u2638-\u263A\u2648-\u2653\u2660\u2663\u2665\u2666\u2668\u267B\u267F\u2692-\u2694\u2696\u2697\u2699\u269B\u269C\u26A0\u26A1\u26AA\u26AB\u26B0\u26B1\u26BD\u26BE\u26C4\u26C5\u26C8\u26CE\u26CF\u26D1\u26D3\u26D4\u26E9\u26EA\u26F0-\u26F5\u26F7-\u26FA\u26FD\u2702\u2705\u2708-\u270D\u270F\u2712\u2714\u2716\u271D\u2721\u2728\u2733\u2734\u2744\u2747\u274C\u274E\u2753-\u2755\u2757\u2763\u2764\u2795-\u2797\u27A1\u27B0\u27BF\u2934\u2935\u2B05-\u2B07\u2B1B\u2B1C\u2B50\u2B55\u3030\u303D\u3297\u3299]|\uD83C[\uDC04\uDCCF\uDD70\uDD71\uDD7E\uDD7F\uDD8E\uDD91-\uDD9A\uDE01\uDE02\uDE1A\uDE2F\uDE32-\uDE3A\uDE50\uDE51\uDF00-\uDF21\uDF24-\uDF93\uDF96\uDF97\uDF99-\uDF9B\uDF9E-\uDFF0\uDFF3-\uDFF5\uDFF7-\uDFFF]|\uD83D[\uDC00-\uDCFD\uDCFF-\uDD3D\uDD49-\uDD4E\uDD50-\uDD67\uDD6F\uDD70\uDD73-\uDD79\uDD87\uDD8A-\uDD8D\uDD90\uDD95\uDD96\uDDA5\uDDA8\uDDB1\uDDB2\uDDBC\uDDC2-\uDDC4\uDDD1-\uDDD3\uDDDC-\uDDDE\uDDE1\uDDE3\uDDEF\uDDF3\uDDFA-\uDE4F\uDE80-\uDEC5\uDECB-\uDED0\uDEE0-\uDEE5\uDEE9\uDEEB\uDEEC\uDEF0\uDEF3]|\uD83E[\uDD10-\uDD18\uDD80-\uDD84\uDDC0]|(?:0\u20E3|1\u20E3|2\u20E3|3\u20E3|4\u20E3|5\u20E3|6\u20E3|7\u20E3|8\u20E3|9\u20E3|#\u20E3|\\*\u20E3|\uD83C(?:\uDDE6\uD83C(?:\uDDEB|\uDDFD|\uDDF1|\uDDF8|\uDDE9|\uDDF4|\uDDEE|\uDDF6|\uDDEC|\uDDF7|\uDDF2|\uDDFC|\uDDE8|\uDDFA|\uDDF9|\uDDFF|\uDDEA)|\uDDE7\uD83C(?:\uDDF8|\uDDED|\uDDE9|\uDDE7|\uDDFE|\uDDEA|\uDDFF|\uDDEF|\uDDF2|\uDDF9|\uDDF4|\uDDE6|\uDDFC|\uDDFB|\uDDF7|\uDDF3|\uDDEC|\uDDEB|\uDDEE|\uDDF6|\uDDF1)|\uDDE8\uD83C(?:\uDDF2|\uDDE6|\uDDFB|\uDDEB|\uDDF1|\uDDF3|\uDDFD|\uDDF5|\uDDE8|\uDDF4|\uDDEC|\uDDE9|\uDDF0|\uDDF7|\uDDEE|\uDDFA|\uDDFC|\uDDFE|\uDDFF|\uDDED)|\uDDE9\uD83C(?:\uDDFF|\uDDF0|\uDDEC|\uDDEF|\uDDF2|\uDDF4|\uDDEA)|\uDDEA\uD83C(?:\uDDE6|\uDDE8|\uDDEC|\uDDF7|\uDDEA|\uDDF9|\uDDFA|\uDDF8|\uDDED)|\uDDEB\uD83C(?:\uDDF0|\uDDF4|\uDDEF|\uDDEE|\uDDF7|\uDDF2)|\uDDEC\uD83C(?:\uDDF6|\uDDEB|\uDDE6|\uDDF2|\uDDEA|\uDDED|\uDDEE|\uDDF7|\uDDF1|\uDDE9|\uDDF5|\uDDFA|\uDDF9|\uDDEC|\uDDF3|\uDDFC|\uDDFE|\uDDF8|\uDDE7)|\uDDED\uD83C(?:\uDDF7|\uDDF9|\uDDF2|\uDDF3|\uDDF0|\uDDFA)|\uDDEE\uD83C(?:\uDDF4|\uDDE8|\uDDF8|\uDDF3|\uDDE9|\uDDF7|\uDDF6|\uDDEA|\uDDF2|\uDDF1|\uDDF9)|\uDDEF\uD83C(?:\uDDF2|\uDDF5|\uDDEA|\uDDF4)|\uDDF0\uD83C(?:\uDDED|\uDDFE|\uDDF2|\uDDFF|\uDDEA|\uDDEE|\uDDFC|\uDDEC|\uDDF5|\uDDF7|\uDDF3)|\uDDF1\uD83C(?:\uDDE6|\uDDFB|\uDDE7|\uDDF8|\uDDF7|\uDDFE|\uDDEE|\uDDF9|\uDDFA|\uDDF0|\uDDE8)|\uDDF2\uD83C(?:\uDDF4|\uDDF0|\uDDEC|\uDDFC|\uDDFE|\uDDFB|\uDDF1|\uDDF9|\uDDED|\uDDF6|\uDDF7|\uDDFA|\uDDFD|\uDDE9|\uDDE8|\uDDF3|\uDDEA|\uDDF8|\uDDE6|\uDDFF|\uDDF2|\uDDF5|\uDDEB)|\uDDF3\uD83C(?:\uDDE6|\uDDF7|\uDDF5|\uDDF1|\uDDE8|\uDDFF|\uDDEE|\uDDEA|\uDDEC|\uDDFA|\uDDEB|\uDDF4)|\uDDF4\uD83C\uDDF2|\uDDF5\uD83C(?:\uDDEB|\uDDF0|\uDDFC|\uDDF8|\uDDE6|\uDDEC|\uDDFE|\uDDEA|\uDDED|\uDDF3|\uDDF1|\uDDF9|\uDDF7|\uDDF2)|\uDDF6\uD83C\uDDE6|\uDDF7\uD83C(?:\uDDEA|\uDDF4|\uDDFA|\uDDFC|\uDDF8)|\uDDF8\uD83C(?:\uDDFB|\uDDF2|\uDDF9|\uDDE6|\uDDF3|\uDDE8|\uDDF1|\uDDEC|\uDDFD|\uDDF0|\uDDEE|\uDDE7|\uDDF4|\uDDF8|\uDDED|\uDDE9|\uDDF7|\uDDEF|\uDDFF|\uDDEA|\uDDFE)|\uDDF9\uD83C(?:\uDDE9|\uDDEB|\uDDFC|\uDDEF|\uDDFF|\uDDED|\uDDF1|\uDDEC|\uDDF0|\uDDF4|\uDDF9|\uDDE6|\uDDF3|\uDDF7|\uDDF2|\uDDE8|\uDDFB)|\uDDFA\uD83C(?:\uDDEC|\uDDE6|\uDDF8|\uDDFE|\uDDF2|\uDDFF)|\uDDFB\uD83C(?:\uDDEC|\uDDE8|\uDDEE|\uDDFA|\uDDE6|\uDDEA|\uDDF3)|\uDDFC\uD83C(?:\uDDF8|\uDDEB)|\uDDFD\uD83C\uDDF0|\uDDFE\uD83C(?:\uDDF9|\uDDEA)|\uDDFF\uD83C(?:\uDDE6|\uDDF2|\uDDFC))))[\ufe00-\ufe0f\u200d]?)+",
          caption_text, re.UNICODE)
      item['tags'] = list(set(item['tags']))

    return item

  def get_original_image(self, url):
    """Gets the full-size image from the specified url."""
    # remove dimensions to get largest image
    url = re.sub(r'/[sp]\d{3,}x\d{3,}/', '/', url)
    # get non-square image if one exists
    url = re.sub(r'/c\d{1,}.\d{1,}.\d{1,}.\d{1,}/', '/', url)
    return url

  def set_story_url(self, item):
    """Sets the story url."""
    urls = []
    if 'video_versions' in item:
      urls.append(item['video_versions'][0]['url'])
    if 'image_versions2' in item:
      urls.append(
          self.get_original_image(item['image_versions2']['candidates'][0][
              'url'].split('?')[0]))
    item['urls'] = urls
    return item

  def download(self, item, save_dir='./'):
    """Downloads the media file."""
    for url in item['urls']:
      base_name = url.split('/')[-1].split('?')[0]
      file_path = os.path.join(save_dir, base_name)

      if not os.path.isfile(file_path):
        with open(file_path, 'wb') as media_file:
          try:
            content = self.session.get(url).content
          except requests.exceptions.ConnectionError:
            time.sleep(5)
            content = self.session.get(url).content

          media_file.write(content)

        timestamp = self.__get_timestamp(item)
        file_time = int(timestamp if timestamp else time.time())
        os.utime(file_path, (file_time, file_time))

  def is_new_media(self, item):
    """Returns True if the media is new."""
    return self.latest is False or self.last_scraped_filemtime == 0 or \
           ('created_time' not in item and 'date' not in item and 'taken_at_timestamp' not in item) or \
           (int(self.__get_timestamp(item)) > self.last_scraped_filemtime)

  @staticmethod
  def __get_timestamp(item):
    return item.get('taken_at_timestamp',
                    item.get('created_time',
                             item.get('taken_at', item.get('date'))))

  @staticmethod
  def __get_file_ext(path):
    return os.path.splitext(path)[1][1:].strip().lower()

  @staticmethod
  def __search(query):
    resp = requests.get(SEARCH_URL.format(query))
    return json.loads(resp.text)

  def search_locations(self):
    query = ' '.join(self.usernames)
    result = self.__search(query)

    if len(result['places']) == 0:
      raise ValueError("No locations found for query '{0}'".format(query))

    sorted_places = sorted(result['places'], key=itemgetter('position'))

    for item in sorted_places[0:5]:
      place = item['place']
      print(
          'location-id: {0}, title: {1}, subtitle: {2}, city: {3}, lat: {4}, lng: {5}'.
          format(place['location']['pk'], place['title'], place['subtitle'],
                 place['location']['city'], place['location']['lat'], place[
                     'location']['lng']))

  @staticmethod
  def save_json(data, dst='./'):
    """Saves the data to a json file."""
    if data:
      with open(dst, 'wb') as f:
        json.dump(
            data,
            codecs.getwriter('utf-8')(f),
            indent=4,
            sort_keys=True,
            ensure_ascii=False)

  @staticmethod
  def get_logger(level=logging.DEBUG, verbose=0):
    """Returns a logger."""
    logger = logging.getLogger(__name__)

    fh = logging.FileHandler('instagram-scraper.log', 'w')
    fh.setFormatter(
        logging.Formatter(
            '%(asctime)s - %(name)s - %(levelname)s - %(message)s'))
    fh.setLevel(level)
    logger.addHandler(fh)

    sh = logging.StreamHandler(sys.stdout)
    sh.setFormatter(logging.Formatter('%(levelname)s: %(message)s'))
    sh_lvls = [logging.ERROR, logging.WARNING, logging.INFO]
    sh.setLevel(sh_lvls[verbose])
    logger.addHandler(sh)

    return logger

  @staticmethod
  def parse_file_usernames(usernames_file):
    """Parses a file containing a list of usernames."""
    users = []

    try:
      with open(usernames_file) as user_file:
        for line in user_file.readlines():
          # Find all usernames delimited by ,; or whitespace
          users += re.findall(r'[^,;\s]+', line.split("#")[0])
    except IOError as err:
      raise ValueError('File not found ' + err)

    return users

  @staticmethod
  def parse_delimited_str(input):
    """Parse the string input as a list of delimited tokens."""
    return re.findall(r'[^,;\s]+', input)


def main():
  parser = argparse.ArgumentParser(
      description="instagram-scraper scrapes and downloads an instagram user's photos and videos.",
      epilog=textwrap.dedent("""
        You can hide your credentials from the history, by reading your
        username from a local file:

        $ instagram-scraper @insta_args.txt user_to_scrape

        with insta_args.txt looking like this:
        -u=my_username
        -p=my_password

        You can add all arguments you want to that file, just remember to have
        one argument per line.

        """),
      formatter_class=argparse.RawDescriptionHelpFormatter,
      fromfile_prefix_chars='@')

  parser.add_argument(
      'username', help='Instagram user(s) to scrape', nargs='*')
  parser.add_argument(
      '--destination', '-d', default='./', help='Download destination')
  parser.add_argument(
      '--login-user',
      '--login_user',
      '-u',
      default=None,
      help='Instagram login user')
  parser.add_argument(
      '--login-pass',
      '--login_pass',
      '-p',
      default=None,
      help='Instagram login password')
  parser.add_argument(
      '--login-only',
      '--login_only',
      '-l',
      default=False,
      action='store_true',
      help='Disable anonymous fallback if login fails')
  parser.add_argument(
      '--filename',
      '-f',
      help='Path to a file containing a list of users to scrape')
  parser.add_argument(
      '--quiet',
      '-q',
      default=False,
      action='store_true',
      help='Be quiet while scraping')
  parser.add_argument(
      '--maximum',
      '-m',
      type=int,
      default=0,
      help='Maximum number of items to scrape')
  parser.add_argument(
      '--retain-username',
      '--retain_username',
      '-n',
      action='store_true',
      default=False,
      help='Creates username subdirectory when destination flag is set')
  parser.add_argument(
      '--media-metadata',
      '--media_metadata',
      action='store_true',
      default=False,
      help='Save media metadata to json file')
  parser.add_argument(
      '--include-location',
      '--include_location',
      action='store_true',
      default=False,
      help='Include location data when saving media metadata')
  parser.add_argument(
      '--media-types',
      '--media_types',
      '-t',
      nargs='+',
      default=['image', 'video', 'story'],
      help='Specify media types to scrape')
  parser.add_argument(
      '--latest',
      action='store_true',
      default=False,
      help='Scrape new media since the last scrape')
  parser.add_argument(
      '--tag',
      action='store_true',
      default=False,
      help='Scrape media using a hashtag')
  parser.add_argument(
      '--filter', default=None, help='Filter by tags in user posts', nargs='*')
  parser.add_argument(
      '--location',
      action='store_true',
      default=False,
      help='Scrape media using a location-id')
  parser.add_argument(
      '--search-location',
      action='store_true',
      default=False,
      help='Search for locations by name')
  parser.add_argument(
      '--comments',
      action='store_true',
      default=False,
      help='Save post comments to json file')
  parser.add_argument(
      '--verbose', '-v', type=int, default=0, help='Logging verbosity level')

  args = parser.parse_args()

  if (args.login_user and args.login_pass is None) or (
      args.login_user is None and args.login_pass):
    parser.print_help()
    raise ValueError('Must provide login user AND password')

  if not args.username and args.filename is None:
    parser.print_help()
    raise ValueError(
        'Must provide username(s) OR a file containing a list of username(s)')
  elif args.username and args.filename:
    parser.print_help()
    raise ValueError(
        'Must provide only one of the following: username(s) OR a filename containing username(s)'
    )

  if args.tag and args.location:
    parser.print_help()
    raise ValueError(
        'Must provide only one of the following: hashtag OR location')

  if args.tag and args.filter:
    parser.print_help()
    raise ValueError('Filters apply to user posts')

  if args.filename:
    args.usernames = InstagramScraper.parse_file_usernames(args.filename)
  else:
    args.usernames = InstagramScraper.parse_delimited_str(','.join(
        args.username))

  if args.media_types and len(args.media_types) == 1 and re.compile(
      r'[,;\s]+').findall(args.media_types[0]):
    args.media_types = InstagramScraper.parse_delimited_str(args.media_types[
        0])

  scraper = InstagramScraper(**vars(args))

  if args.tag:
    scraper.scrape_hashtag()
  elif args.location:
    scraper.scrape_location()
  elif args.search_location:
    scraper.search_locations()
  else:
    scraper.scrape()


if __name__ == '__main__':
  main()



"""Scraper/api for imdb.
"""

from imdb import IMDb

from datahunters.movies import common


class IMDBAPI(common.MovieAPIBase):
  """Class for IMDB.
  """
  engine = NotImplemented

  def __init__(self, api_key=None):
    super(IMDBAPI, self).__init__(api_key)
    self.engine = IMDb()
    print("init imdb api")

  def get_movie_info(self, id):
    data_mat = self.engine.get_movie(id)
    # print data_mat["full-size cover url"]
    # print data_mat["genres"]
    # print data_mat["cast"]

  def get_person_info(self, name):
    pass

  def map_to_internal_genre(self, genre):
    pass


if __name__ == "__main__":
  api = IMDBAPI()
  api.get_movie_info("3170832")



"""Web driver downloader.
"""

import os
import platform
import urllib
import stat
import tarfile
import zipfile


def download_chrome_driver():
  linux_url = "https://chromedriver.storage.googleapis.com/2.34/chromedriver_linux64.zip"
  mac_url = "https://chromedriver.storage.googleapis.com/2.34/chromedriver_mac64.zip"
  win_url = "https://chromedriver.storage.googleapis.com/2.34/chromedriver_win32.zip"
  cur_dir = os.path.dirname(os.path.realpath(__file__))
  save_dir = os.path.join(cur_dir, "chrome")
  if not os.path.exists(save_dir):
    os.makedirs(save_dir)
  data_fn = os.path.join(save_dir, "driver.zip")
  if platform.system() == "Linux":
    download_url = linux_url
  elif platform.system() == "Darwin":
    download_url = mac_url
  else:
    download_url = win_url
  print("downloading chrome driver, this may take some time...")
  filepath, _ = urllib.request.urlretrieve(download_url, data_fn)
  statinfo = os.stat(filepath)
  print("Successfully downloaded", data_fn, statinfo.st_size, "bytes.")
  print("unzipping...")
  zip_ref = zipfile.ZipFile(data_fn, "r")
  zip_ref.extractall(save_dir)
  zip_ref.close()
  os.remove(data_fn)
  # change permission.
  driver_fn = os.path.join(save_dir, "chromedriver")
  st = os.stat(driver_fn)
  os.chmod(driver_fn, st.st_mode | stat.S_IEXEC)
  print("finished.")


def download_phantom_driver():
  linux_url = "https://bitbucket.org/ariya/phantomjs/downloads/phantomjs-2.1.1-linux-x86_64.tar.bz2"
  mac_url = "https://bitbucket.org/ariya/phantomjs/downloads/phantomjs-2.1.1-macosx.zip"
  win_url = "https://bitbucket.org/ariya/phantomjs/downloads/phantomjs-2.1.1-windows.zip"
  cur_dir = os.path.dirname(os.path.realpath(__file__))
  save_dir = os.path.join(cur_dir, "phantom")
  if not os.path.exists(save_dir):
    os.makedirs(save_dir)
  if platform.system() == "Linux":
    data_fn = os.path.join(save_dir, "driver.tar.bz2")
    download_url = linux_url
  else:
    data_fn = os.path.join(save_dir, "driver.zip")
    download_url = mac_url
  print("downloading phantom driver, this may take some time...")
  filepath, _ = urllib.request.urlretrieve(download_url, data_fn)
  statinfo = os.stat(filepath)
  print("Successfully downloaded", data_fn, statinfo.st_size, "bytes.")
  print("unzipping...")
  if platform.system() == "Linux":
    tarfile.open(data_fn, 'r:bz2').extractall(save_dir)
  else:
    zip_ref = zipfile.ZipFile(data_fn, "r")
    zip_ref.extractall(save_dir)
    zip_ref.close()
  os.remove(data_fn)
  print("finished.")


def download_all_drivers():
  download_chrome_driver()
  download_phantom_driver()


if __name__ == "__main__":
  # download_chrome_driver()
  download_phantom_driver()



"""Facebook graph api.
"""

import json
import requests

from datahunters.social.common import SocialUser
from datahunters.social.tools import gen_fb_token, gen_fb_token_nonexpired, unicode_normalize


class FBPost(object):
  """Class for facebook post.
  """

  def __init__(self):
    self.owner = ""

  def to_json(self):
    return json.loads(json.dumps(self.__dict__))


class FacebookAPI(object):
  """Class for Facebook graph api.
  """
  GRAPH_API_URL = "https://graph.facebook.com/"

  def __init__(self, fb_email=None, fb_pw=None, fb_token=None):
    if fb_token is None:
      assert fb_email is not None and fb_pw is not None
      self.fb_token = gen_fb_token(fb_email, fb_pw)
    else:
      self.fb_token = fb_token
    self.VERSION = "v2.12"
    self.test = None
    # subject specific property.
    self.page_fields = []
    self.post_fields = [
        "name", "id", "caption", "created_time", "description", "from",
        "full_picture", "link", "place", "type", "updated_time",
        "permalink_url", "shares",
        "reactions.type(LIKE).limit(0).summary(total_count).as(like)",
        "reactions.type(LOVE).limit(0).summary(total_count).as(love)",
        "reactions.type(WOW).limit(0).summary(total_count).as(wow)",
        "reactions.type(HAHA).limit(0).summary(total_count).as(haha)",
        "reactions.type(SAD).limit(0).summary(total_count).as(sad)",
        "reactions.type(ANGRY).limit(0).summary(total_count).as(angry)",
        "reactions.limit(0).summary(true)", "comments.limit(0).summary(true)"
    ]
    self.group_fields = ["cover", "email", "description"]
    self.user_fields = []

  """General graph api.
  """

  def get_object(self, obj_id, fields):
    """Retrieve graph object.
    """
    field_str = ""
    for field in fields:
      field_str += "{},".format(field)
    field_str = field_str.rstrip(",")
    url = "{}{}/{}?access_token={}&fields={}".format(
        self.GRAPH_API_URL, self.VERSION, obj_id, self.fb_token, field_str)
    res = requests.get(url).json()
    return res

  def get_edge(self, obj_id, edge, fields):
    """Retrieve edge related to an object.
    """
    field_str = ""
    for field in fields:
      field_str += "{},".format(field)
    field_str = field_str.rstrip(",")
    url = "{}{}/{}/{}?access_token={}&fields={}".format(
        self.GRAPH_API_URL, self.VERSION, obj_id, edge, self.fb_token,
        field_str)
    res = requests.get(url).json()
    return res

  def build_search_url(self, search_type, query):
    """Create a url for search request.
    """
    assert search_type in ["group", "user", "event", "place", "page"]
    url = "{}{}/search?access_token={}&q={}&type={}&limit=25".format(
        self.GRAPH_API_URL, self.VERSION, self.fb_token,
        query.replace(" ", "+"), search_type)
    return url

  """Group related code.
  """

  def search_groups(self, query):
    """Search groups and use as generator.
    """
    next_url = self.build_search_url("group", query)
    while next_url is not None:
      res = requests.get(next_url).json()
      if "next" in res["paging"]:
        next_url = res["paging"]["next"]
      else:
        next_url = None
      for group in res["data"]:
        # only public fields.
        group_obj = self.get_object(group["id"], self.group_fields)
        yield group_obj
    yield None

  def get_group(self, group_id):
    """Retrieve group object.
    """
    pass

  """User related code.
  """

  def search_users(self, query):
    """Search users.
    """
    next_url = self.build_search_url("user", query)
    while next_url is not None:
      res = requests.get(next_url).json()
      if "next" in res["paging"]:
        next_url = res["paging"]["next"]
      else:
        next_url = None
      for user in res["data"]:
        # only public fields.
        user_obj = self.get_object(user["id"], [])
        yield user_obj
    yield None

  def get_user(self, user_id):
    """Retrieve user information.
    """
    user = self.engine.get("/{}".format(user_id))
    return user

  def get_my_post(self, post_id):
    pass

  """Event related code.
  """

  def search_events(self, query):
    """Search events.
    """
    event_gen = self.engine.search(query, "event", True)

  """Page related code.
  """

  def search_pages(self, query):
    """Search pages.
    """
    next_url = self.build_search_url("page", query)
    while next_url is not None:
      res = requests.get(next_url).json()
      if "next" in res["paging"]:
        next_url = res["paging"]["next"]
      else:
        next_url = None
      for group in res["data"]:
        # only public fields.
        group_obj = self.get_object(group["id"],
                                    ["cover", "email", "description"])
        yield group_obj
    yield None

  def get_page_info(self, page_id):
    """Retrieve basic info about page.
    """
    pass

  def get_page_posts(self, page_id):
    """Retrieve all posts from a page.

    Args:
      page_id: name of the page from url.
    """
    res = self.get_edge(page_id, "posts", self.post_fields)
    next_url = "start"
    while next_url is not None:
      # get next url.
      if "paging" in res.keys():
        next_url = res["paging"]["next"]
      else:
        next_url = None
      # process posts.
      posts = res["data"]
      for post in posts:
        yield post

  """Post related code.
  """

  def get_post(self, post_id):
    """Retrieve info about post.
    """
    post = self.get_object(post_id, self.post_fields)
    return post


if __name__ == "__main__":
  token = gen_fb_token_nonexpired("2141139379235357",
                                  "6b1316b647bef3f306453a38d1b9e380")
  api = FacebookAPI(fb_token=token)
  # print api.__dict__
  post_gen = api.get_page_posts("BetterHQ")
  post = next(post_gen)
  print post



"""Base api class.
"""

import abc
import json


class StockContent(object):
  """Class for a single stock content object.

  Example content: image, video, 3d model.
  """
  img_id = ""
  title = ""
  description = ""
  owner = ""
  created_date = ""
  full_url = ""
  normal_url = ""
  link = ""


class StockImageAPIBase(object):
  """Base class for stock image apis.
  """
  __metaclass__ = abc.ABCMeta

  def save_json_data(self, json_data, save_fn):
    """Save json to file.
    """
    with open(save_fn, "w") as f:
      json.dump(json_data, f)

  @abc.abstractmethod
  def search_imgs(self, keywords):
    """Search images using keywords.

    Args:
      keywords: descriptions of the topic.
      num: number of images to return.

    Returns:
      generator for continuously fetching images until None.
    """
    pass

  @abc.abstractmethod
  def get_imgs(self, start_page=-1):
    """Fetch images from the whole collection.

    Ordered by latest.

    Args:
      start_page: page number to start.

    Returns:
       generator for continuously fetching images until None.
    """
    pass



"""Tests for common.
"""

import json

from datahunters.movies import common


class DummyMovieAPI(common.MovieAPIBase):
  def __init__(self, api_key):
    super(DummyMovieAPI, self).__init__(api_key)

  def get_movie_info(self, id):
    pass

  def map_to_internal_genre(self, genre):
    pass

  def get_person_info(self, id):
    pass


class TestMovieCommon(object):
  def test_genre_listing(self):
    api = DummyMovieAPI(None)
    genres = api.list_internal_genres()
    print genres